@@ -1,7 +1,125 @@
+
+#' Estimation of limited dependent variable models
+#' 
+#' mhurdle fits a large set of models relevant when the dependent variable is 0
+#' for a part of the sample.
+#' 
+#' 
+#' `mhurdle` fits models for which the dependent variable is zero for
+#' a part of the sample. Null values of the dependent variable may
+#' occurs because of one or several mechanisms : good rejection, lack
+#' of ressources and purchase infrequency. The model is described
+#' using a three-parts formula : the first part describes the
+#' selection process if any, the second part the regression equation
+#' and the third part the purchase infrequency process.  `y ~ 0 | x1 +
+#' x2 | z1 + z2` means that there is no selection process.  `y ~ w1 +
+#' w2 | x1 + x2 | 0` and `y ~ w1 + w2 | x1 + x2` describe the same
+#' model with no purchase infrequency process. The second part is
+#' mandatory, it explains the positive values of the dependant
+#' variable. The `dist` argument indicates the distribution of the
+#' error term. If `dist = "n"`, the error term is normal and (at least
+#' part of) the zero observations are also explained by the second
+#' part as the result of a corner solution. Several models described
+#' in the litterature are obtained as special cases :
+#' 
+#' A model with a formula like `y~0|x1+x2` and `dist="n"` is the Tobit
+#' model proposed by \insertCite{TOBIN/58}{mhurdle}.
+#' 
+#' `y~w1+w2|x1+x2` and `dist="l"` or `dist="t"` is the single hurdle
+#' model proposed by \insertCite{CRAGG/71}{mhurdle}. With `dist="n"`,
+#' the double hurdle model also proposed by
+#' \insertCite{CRAGG/71}{mhurdle} is obtained. With `corr="h1"` we get
+#' the correlated version of this model described by
+#' \insertCite{BLUNDELL/87}{mhurdle}.
+#' 
+#' `y~0|x1+x2|z1+z2` is the P-Tobit model of
+#' \insertCite{DEATO/IRISH/84}{mhurdle}, which can be a single hurdle
+#' model if `dist="t"` or `dist="l"` or a double hurdle model if
+#' `dist="n"`.
+#' 
+#' @name mhurdle
+#' @aliases mhurdle
+#' @param formula a symbolic description of the model to be fitted,
+#' @param data a `data.frame`,
+#' @param subset see [stats::lm()],
+#' @param weights see [stats::lm()],
+#' @param na.action see [stats::lm()],
+#' @param start starting values,
+#' @param dist the distribution of the error of the consumption
+#'     equation: one of `"n"` (normal), `"ln"` (log-normal) `"bc"`
+#'     (box-cox normal) and `"ihs"` (inverse hyperbolic sinus
+#'     transformation),
+#' @param h2 if `TRUE` the second hurdle is effective, it is not
+#'     otherwise,
+#' @param scaled if `TRUE`, the dependent variable is divided by its
+#'     geometric mean,
+#' @param corr a boolean indicating whether the errors of the
+#'     different equations are correlated or not,
+#' @param robust transformation of the structural parameters in order
+#'     to avoid numerical problems,
+#' @param check_gradient if `TRUE`, a matrix containing the analytical and
+#'     the numerical gradient for the starting values are returned,
+#' @param \dots further arguments.
+#' @return
+#' #' an object of class `c("mhurdle", "maxLik")`.
+#' 
+#' A `mhurdle` object has the following elements :
+#' 
+#' - coefficients: the vector of coefficients,
+#' - vcov: the covariance matrix of the coefficients,
+#' - fitted.values: a matrix of fitted.values, the first column being
+#' the probability of 0 and the second one the mean values for the
+#' positive observations,
+#' - logLik: the log-likelihood,
+#' - gradient: the gradient at convergence,
+#' - model: a data.frame containing the variables used for the estimation,
+#' - coef.names: a list containing the names of the coefficients in the
+#' selection equation, the regression equation, the infrequency of purchase
+#' equation and the other coefficients (the standard deviation of the error
+#' term and the coefficient of correlation if `corr = TRUE`,
+#' - formula: the model formula, an object of class `Formula`
+#' - call: the call,
+#' - rho: the lagrange multiplier test of no correlation.
+#' 
+#' @references
+#'
+#' \insertRef{BLUNDELL/87}{mhurdle}
+#'
+#' \insertRef{CRAGG/71}{mhurdle}
+#'
+#' \insertRef{DEATO/IRISH/84}{mhurdle}
+#'
+#' \insertRef{TOBIN/58}{mhurdle}
+#'
+#' @keywords regression
+#' @examples
+#' 
+#' data("Interview", package = "mhurdle")
+#' 
+#' # independent double hurdle model
+#' idhm <- mhurdle(vacations ~ car + size | linc + linc2 | 0, Interview,
+#'               dist = "ln", h2 = TRUE, method = "bfgs")
+#' 
+#' # dependent double hurdle model
+#' ddhm <- mhurdle(vacations ~ car + size | linc + linc2  | 0, Interview,
+#'               dist = "ln", h2 = TRUE, method = "bfgs", corr = TRUE)
+#' 
+#' # a double hurdle p-tobit model
+#' ptm <- mhurdle(vacations ~ 0 | linc + linc2 | car + size, Interview,
+#'               dist = "ln", h2 = TRUE, method = "bfgs", corr = TRUE)
+#' @importFrom Formula Formula
+#' @importFrom survival survreg Surv
+#' @importFrom truncreg truncreg
+#' @importFrom stats binomial cor df.residual dnorm ecdf formula glm
+#'     integrate lm lm.fit model.frame model.matrix model.response
+#'     optimize pchisq pnorm printCoefmat qnorm quantile rnorm terms
+#'     uniroot var .getXlevels deviance
+#' @importFrom maxLik maxLik activePar
+#' @export
 mhurdle <- function(formula, data, subset, weights, na.action,
                     start = NULL, dist = c("ln", "n", "bc", "ihs"), h2 = FALSE,
                     scaled = TRUE, corr = FALSE, robust = TRUE,
-                    check.grad = FALSE, ...){
+                    check_gradient = FALSE, ...){
     fitted = TRUE
     dots <- list(...)
     oldoptions <- options(warn = -1)
@@ -14,7 +132,6 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     if (dist == "ln" & h2) dist <- "ln2"
     if (dist == "n" & ! h2) dist <- "tn"
     if (dist == "bc" & h2) dist <- "bc2"
-    
     # 1. Compute the model.frame and the model.matrix
 
     if (! inherits(formula, "Formula")) formula <- Formula(formula)
@@ -29,6 +146,7 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     mf[[1L]] <- as.name("model.frame")
     mf$formula <- formula
     mf <- eval(mf, parent.frame())
+    .terms <- terms(mf)
     X1 <- model.matrix(formula, data = mf, rhs = 1)
     X2 <- model.matrix(formula, data = mf, rhs = 2)
     X3 <- model.matrix(formula, data = mf, rhs = 3)
@@ -42,6 +160,7 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     if (scaled){
         geomean <- exp(mean(log(y[y > 0])))
         y <- y / geomean
+        attr(y, "geomean") <- attr(mf, "geomean") <- geomean
     }
     N <- length(y)
     if (length(X1) == 0) X1 <- NULL
@@ -52,12 +171,17 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     h3 <- ! is.null(X3)
     
     #  2. One equation models
-    if (! h1 && ! h3){
-        result <- onequation.mhurdle(X2, y, dist)
+    if (! h1 && ! h3 && is.null(X4)){
+        result <- one_equation_model(X2, y, dist, check_gradient = check_gradient, start = start)
+        if (check_gradient){
+            result$call <- cl
+            return(result)
+        }
         result$naive <- NULL#naive
         result$call <- cl
         result$model <- mf
         result$formula <- formula
+        result$terms <- .terms
         result$R2 <- c(null = NA, positive = NA)
         return(result)
     }
@@ -65,12 +189,17 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     # 3. Selection single hurdle models without correlation that can
     # be estimated simply in two parts using seperate.mhurdle()
     if (h1 && !h3  && ! corr && !h2){
-        result <- seperate.mhurdle(X1, X2, y, dist = dist)
+        result <- two_parts_model(X1, X2, y, dist = dist, check_gradient = check_gradient, start = start)
+        if (check_gradient){
+            result$call <- cl
+            return(result)
+        }
         result$naive <- NULL#naive
         result$call <- cl
         result$model <- mf
         result$formula <- formula
         result$R2 <- c(null = NA, positive = NA)
+        result$terms <- .terms
         return(result)
     }
 
@@ -83,7 +212,6 @@ mhurdle <- function(formula, data, subset, weights, na.action,
         if (dist %in% c("bc", "bc2", "ln2")) dist.start <- "ln"
         if (dist == "ihs") dist.start <- "n"
         start <- start.mhurdle(X1, X2, X3, y, dist.start)
-        
         # in case of heteroscedasctic model, add K4 zeros to the start
         # vector and the intercept should be ln(sigma_o) (not sigma_o)
         # because of the exp form
@@ -95,26 +223,35 @@ mhurdle <- function(formula, data, subset, weights, na.action,
         if (corr){
             if (robust) rhoinit <- tan(0.1 * pi / 2) else rhoinit <- 0.1
             if (h1 + h3 == 2) start <- c(start, rho12 = rhoinit, rho13 = rhoinit, rho23 = rhoinit)
-            else start <- c(start, rhoinit)
+            else start <- c(start, rho = rhoinit)
+        }
+        if (dist %in% c("bc", "bc2", "ihs")) start <- c(start, tr =  0.1)
+        if (dist %in% c("bc2", "ln2")){
+            pos = 1
+            if (robust) pos <- max(log(pos), -100)
+            start <- c(start, pos = pos)
         }
-        if (dist %in% c("bc", "bc2", "ihs")) start <- c(start, tr = -0.1)
-        if (dist %in% c("bc2", "ln2")) start <- c(start, pos = 1)
     }
     else{
         if (robust){
             sd.pos <- getindex(X1, X2, X3, X4, corr, dist, which = "sd")
-            mu.pos <- getindex(X1, X2, X3, X4, corr, dist, which = "pos")
+            if (dist %in% c("ln2", "bc2")) mu.pos <- getindex(X1, X2, X3, X4, corr, dist, which = "pos")
             rho.pos <- getindex(X1, X2, X3, X4, corr, dist, which = "corr")
-            start[c(sd.pos, mu.pos)] <- log(start[c(sd.pos, mu.pos)])
-            start[rho.pos] <- tan(start[rho.pos] * pi / 2)
+            start[sd.pos] <- log(start[sd.pos])
+            if (dist %in% c("ln2", "bc2")) start[mu.pos] <- max(log(start[mu.pos]), - 50)
+            if (corr) start[rho.pos] <- tan(start[rho.pos] * pi / 2)
         }
     }
+
     result <- mhurdle.fit(start, X1, X2, X3, X4, y, gradient = TRUE,
                           dist = dist, corr = corr, robust = robust,
-                          fitted = fitted, check.grad = check.grad,
+                          fitted = fitted,
+                          check_gradient = check_gradient,
                           ...)
-    if (check.grad) return(result)
-    if (fitted & scaled) result$fitted.values[, 2] <- result$fitted.values[, 2] * geomean
+    if (check_gradient){
+        result$call <- cl
+        return(result)
+    }
     
     # 3. Compute the naive model
 
@@ -132,7 +269,7 @@ mhurdle <- function(formula, data, subset, weights, na.action,
 
     start.naive <- c(rep(0.1, 1 + h1 + h3), 1)
     moments <- c(Pnull, Ec, Vc)
-    naive <- maxLik(lnl.naive, start = start.naive,
+    naive <- maxLik::maxLik(lnl.naive, start = start.naive,
                     dist = dist.naive, moments = moments,
                     h1 = h1, h3 = h3)
     coef.naive <- naive$est
@@ -146,6 +283,8 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     result$call <- cl
     result$formula <- formula
     result$model <- mf
+    result$terms <- .terms
+    result$xlevels = .getXlevels(.terms, mf)
     
     lnL1c <- attr( naive$logLik, "parts")[1] + attr( naive$logLik, "parts")[2]
     lnL1u <- attr(result$logLik, "parts")[1] + attr(result$logLik, "parts")[2]
@@ -154,16 +293,17 @@ mhurdle <- function(formula, data, subset, weights, na.action,
     lnL2u <- attr(result$logLik, "parts")[3] - attr(result$logLik, "parts")[2]
 
     R1 <- 1 - (lnL1c / lnL1u) ^ (2 / N * lnL1c)
-    R2 <- 1 - (exp(lnL2c) / exp(lnL2u)) ^ (2 / (N * (1 - Pnull)))
-    
-    result$R2 <- c(null = R1, positive = R2)
+    R2old <- 1 - (exp(lnL2c) / exp(lnL2u)) ^ (2 / (N * (1 - Pnull)))
+    R2 <- 1 - exp(- 2 * (lnL2u - lnL2c) / (N * (1 - Pnull)))
+    result$R2 <- c(null = unname(R1), positive = unname(R2), old = unname(R2old))
+    result$dpar <- - sum(diag(solve(result$hessian, crossprod(result$gradient))))
     result
 }
 
 mhurdle.fit <- function(start, X1, X2, X3, X4, y, gradient = FALSE, fit = FALSE,
                         dist = c("ln", "n", "tn", "bc", "ihs", "bc2", "ln2"),
                         corr = FALSE, robust = TRUE,  fitted = FALSE,
-                        check.grad = FALSE, ...){
+                        check_gradient = FALSE, ...){
     start.time <- proc.time()
     h1 <- ! is.null(X1)
     h3 <- ! is.null(X3)
@@ -195,25 +335,14 @@ mhurdle.fit <- function(start, X1, X2, X3, X4, y, gradient = FALSE, fit = FALSE,
     if (h3) start.names$h3 <- paste("h3", start.names$h3, sep = ".")
     if (h4) start.names$h4 <- paste("h4", start.names$h4, sep = ".")
     names(start) <- Reduce("c", start.names)
-
     f <- function(param) mhurdle.lnl(param, X1 = X1, X2 = X2, X3 = X3, X4 = X4, y = y,
                                      gradient = TRUE, fitted = FALSE,
                                      dist = dist, corr = corr,
                                      robust = robust)
-    if (check.grad){
-        ngrad <- c()
-        oparam <- start
-        fo <- f(start)
-        agrad <- apply(attr(fo, "gradient"), 2, sum)
-        eps <- 1E-05
-        for (i in 1:length(start)){
-            oparam[i] <- oparam[i] + eps
-            ngrad <- c(ngrad, sum((as.numeric(f(oparam)) - fo) / eps))
-            oparam <- start
-        }
-        return(cbind(start, agrad, ngrad))
-    }
-    maxl <- maxLik(f, start = start, control = list(lambdatol = 1E-20),  ...)
+
+    if (check_gradient) return(compare_gradient(f, start))
+
+    maxl <- maxLik::maxLik(f, start = start, control = list(lambdatol = 1E-20),  ...)
     nb.iter <- maxl$iterations
     convergence.OK <- maxl$code <= 2
     coefficients <- maxl$estimate
@@ -231,8 +360,14 @@ mhurdle.fit <- function(start, X1, X2, X3, X4, y, gradient = FALSE, fit = FALSE,
                         parts = attr(logLik, "parts"),
                         nobs = length(y), class = "logLik")
     hessian <- maxl$hessian
+    ev <- eigen(- hessian)$values
+    if (any(ev < 0)) cat("The hessian is not negative definite\n")
+    else (if (any(abs(ev) < 1E-07)) cat("the hessian is singular\n"))
+    opg <- crossprod(gradi)
     elaps.time <- proc.time() - start.time
-    eps <- with(maxl, gradient %*% solve(- hessian) %*% gradient)
+    
+#    eps <- with(maxl, gradient %*% solve(- hessian) %*% gradient)
+    eps <- with(maxl, gradient %*% solve(opg) %*% gradient)
     est.stat <- list(elaps.time = elaps.time,
                      nb.iter = nb.iter,
                      eps = eps,
@@ -257,12 +392,13 @@ mhurdle.fit <- function(start, X1, X2, X3, X4, y, gradient = FALSE, fit = FALSE,
         coefficients[possd] <- exp(coefficients[possd])
     }
     result <- list(coefficients  = coefficients,
-                   vcov          = diag(gtheta) %*% (- solve(maxl$hessian) ) %*% diag(gtheta),
                    fitted.values = fitted.values,
                    logLik        = logLik,
                    gradient      = gradi,
                    hessian       = hessian,
                    formula       = NULL,
+                   terms         = NULL,
+                   xlevels       = NULL,
                    model         = NULL,
                    coef.names    = coef.names,
                    call          = NULL,
